/*
Website RSS generator.

The program accepts a number of document meta files on the command line. Each
such file contain information about a document of the same file name except for
a html extension. In addition to this, the program expects to find certain site
variables in the environment. These are documented further down. A document link
is constructed based on the location of the meta file and the site_link
variable. If an index.html file is found pointing to the calculated document
link, the final document link is made as the directory of said index.html file.

A meta file is a shell script containing string variables.

    - title *
    - description *
    - lang
    - topics
    - published *
    - revised

Meta variables with an asterisk are mandatory. The variable 'topics' contain a
list of words (hash tags) that categorize the document content. The variable
'description' is parsed as markdown.

Example meta file:

    title="Expired Efke R50 Test Roll"
    topics="Photo"
    lang=en
    published="2024-01-02"
    revised="2024-01-03"
    description="Fotokemika company background and experience developing and scanning an expired roll of Efke R50 film."

The following site variables are recognized by the program:

    - site_link *
    - site_title *
    - site_description *
    - site_lang
    - site_webmaster

Site variables with an asterisk are mandatory.

Example site variables:

    site_link="https://bitmaster.se"
    site_title="Bitmaster(se)"
    site_description="Hobbies and ramblings of mine"
    site_lang=en
    site_webmaster="zero@dev.null"

The program outputs the generated RSS file to stdout with a zero exit code.

Any error is reported to stderr with a non-zero exit code.
*/

package main

import (
	"bytes"
	"context"
	"encoding/xml"
	"fmt"
	"net/url"
	"os"
	"path/filepath"
	"slices"
	"strings"
	"time"

	"mvdan.cc/sh/v3/interp"
	"mvdan.cc/sh/v3/syntax"

	"github.com/gomarkdown/markdown"
	"github.com/gomarkdown/markdown/html"
	"github.com/gomarkdown/markdown/parser"
)

type RSS struct {
	XMLName xml.Name `xml:"rss"`
	Version string   `xml:"version,attr"`
	Channel Channel  `xml:"channel"`
}

type Channel struct {
	SiteInfo
	LastBuildDate XMLTime        `xml:"lastBuildDate"`
	Documents     []DocumentInfo `xml:"item"`
}

type SiteInfo struct {
	Title          string `xml:"title"`
	Link           string `xml:"link"`
	Description    string `xml:"description"`
	Language       string `xml:"language,omitempty"`
	WebmasterEmail string `xml:"webMaster,omitempty"`
}

type DocumentInfo struct {
	Title       string `xml:"title"`
	Link        string `xml:"link"`
	Guid        string `xml:"guid"`
	Description struct {
		D string `xml:",cdata"`
	} `xml:"description"`
	Language      string    `xml:"-"`
	Topics        []string  `xml:"-"`
	PublishedDate time.Time `xml:"pubDate"`
	RevisedDate   time.Time `xml:"-"`
}

type XMLTime struct {
	time.Time
}

func (t *XMLTime) MarshalXML(e *xml.Encoder, start xml.StartElement) error {
	return e.EncodeElement(t.Format(time.RFC822), start)
}

func main() {
	siteInfo, err := GetSiteInfo()
	if err != nil {
		fatalln(err)
	}
	var documents []DocumentInfo
	for _, metaPath := range os.Args[1:] {
		docInfo, err := GetDocumentInfo(metaPath, &siteInfo)
		if err != nil {
			fatalf("Failed to load meta file '%s': %v\n", metaPath, err)
		}
		documents = append(documents, docInfo)
	}

	slices.SortFunc(documents, func(a, b DocumentInfo) int {
		if order := a.PublishedDate.Compare(b.PublishedDate); order != 0 {
			return order
		}
		return strings.Compare(a.Link, b.Link)
	})
	slices.Reverse(documents)

	rss := RSS{
		Version: "2.0",
		Channel: Channel{
			SiteInfo: siteInfo,
			LastBuildDate: XMLTime{
				Time: time.Now(),
			},
			Documents: documents,
		},
	}

	var buf bytes.Buffer
	buf.WriteString(`<?xml version="1.0" encoding="UTF-8"?>` + "\n")

	enc := xml.NewEncoder(&buf)
	enc.Indent("", "  ")
	if err = enc.Encode(&rss); err != nil {
		fatalf("xml encode error: %v\n", err)
	}
	buf.WriteString("\n")
	buf.WriteTo(os.Stdout)
}

func GetSiteInfo() (si SiteInfo, err error) {
	getMandatoryVar := func(name string, value *string, err *error) bool {
		if *value = os.Getenv(name); *value == "" {
			*err = fmt.Errorf("missing environment variable '%s'", name)
			return false
		}
		return true
	}
	if !getMandatoryVar("site_link", &si.Link, &err) ||
		!getMandatoryVar("site_title", &si.Title, &err) ||
		!getMandatoryVar("site_description", &si.Description, &err) {
		return SiteInfo{}, err
	}
	si.Language = os.Getenv("site_lang")
	si.WebmasterEmail = os.Getenv("site_webmaster")
	return
}

func GetDocumentInfo(metaPath string, si *SiteInfo) (di DocumentInfo, err error) {
	metaFile, err := os.Open(metaPath)
	if err != nil {
		return DocumentInfo{}, err
	}
	defer metaFile.Close()

	shRunner, err := interp.New(interp.StdIO(strings.NewReader(""), nil, nil))
	if err != nil {
		return DocumentInfo{}, err
	}

	shParser := syntax.NewParser()

	shProg, err := shParser.Parse(metaFile, "")
	if err != nil {
		return DocumentInfo{}, fmt.Errorf("script parse error: %v\n", err)
	}

	if err = shRunner.Run(context.Background(), shProg); err != nil {
		return DocumentInfo{}, fmt.Errorf("script eval error: %v\n", err)
	}

	getMandatoryVar := func(name string, value *string, err *error) bool {
		if *value = shRunner.Vars[name].String(); *value == "" {
			*err = fmt.Errorf("missing variable '%s'", name)
			return false
		}
		return true
	}

	var description, publishedDate string

	if !getMandatoryVar("title", &di.Title, &err) ||
		!getMandatoryVar("description", &description, &err) ||
		!getMandatoryVar("published", &publishedDate, &err) {
		return DocumentInfo{}, err
	}

	di.Description.D = "\n" + string(htmlOfMarkdown([]byte(description)))
	di.Language = shRunner.Vars["lang"].String()
	di.Topics = strings.Fields(shRunner.Vars["topics"].String())

	revisedDate := shRunner.Vars["revised"].String()

	convertDate := func(metaDate string) (time.Time, error) {
		if metaDate = strings.TrimSpace(metaDate); metaDate == "" {
			return time.Time{}, nil
		}
		return time.ParseInLocation(time.DateOnly, metaDate, time.Local)
	}

	if di.PublishedDate, err = convertDate(publishedDate); err != nil {
		return DocumentInfo{}, fmt.Errorf("published date parse error: %v", err)
	}
	if di.PublishedDate.IsZero() {
		return DocumentInfo{}, fmt.Errorf("published date is empty!")
	}
	if di.RevisedDate, err = convertDate(revisedDate); err != nil {
		return DocumentInfo{}, fmt.Errorf("revised date parse error: %v", err)
	}

	htmlPath := strings.TrimSuffix(metaPath, filepath.Ext(metaPath)) + ".html"
	if _, err = os.Stat(htmlPath); err != nil {
		return DocumentInfo{}, fmt.Errorf("missing derived document: %v", err)
	}
	if di.Link, err = url.JoinPath(si.Link, htmlPath); err != nil {
		return DocumentInfo{}, fmt.Errorf("URL construction error: %v", err)
	}

	htmlName := filepath.Base(htmlPath)
	htmlDir := filepath.Dir(htmlPath)
	indexPath := filepath.Join(htmlDir, "index.html")

	if fi, err := os.Lstat(indexPath); err == nil && fi.Mode()&os.ModeSymlink != 0 {
		if name, _ := os.Readlink(indexPath); name == htmlName {
			// Link to directory containing docuement as there is an "index.html".
			if di.Link, err = url.JoinPath(si.Link, htmlDir); err != nil {
				return DocumentInfo{}, fmt.Errorf("URL construction error: %v", err)
			}
		}
	}
	// Guid is the unique identifier of an item.
	// Use the link as the guid as we aim not to rename links.
	di.Guid = di.Link

	return di, nil
}

func htmlOfMarkdown(md []byte) []byte {
	extensions := parser.CommonExtensions | parser.AutoHeadingIDs | parser.NoEmptyLineBeforeBlock
	ast := parser.NewWithExtensions(extensions).Parse(md)
	htmlFlags := html.CommonFlags | html.HrefTargetBlank
	htmlRenderer := html.NewRenderer(html.RendererOptions{Flags: htmlFlags})
	return markdown.Render(ast, htmlRenderer)
}

func fatalf(format string, a ...interface{}) {
	fmt.Fprintf(os.Stderr, format, a...)
	os.Exit(1)
}

func fatalln(a ...interface{}) {
	fmt.Fprintln(os.Stderr, a...)
	os.Exit(1)
}
